@@ -24,6 +24,6 @@
 #
 ###############################################################################
 
-__version__ = '22.2.2'
+__version__ = '22.4.2'
 
-__build__ = u'00000000-0000000'
+__build__ = '00000000-0000000'
