@@ -72,7 +72,7 @@ def _validate_endpoint(endpoint, check_native_endpoint=None):
             raise ValueError('invalid type "{}" in endpoint'.format(endpoint['type']))
 
         for k in endpoint.keys():
-            if k not in ['type', 'host', 'port', 'path', 'tls']:
+            if k not in ['type', 'host', 'port', 'path', 'tls', 'timeout', 'version']:
                 raise ValueError(
                     "Invalid key '{}' in endpoint configuration".format(k)
                 )
@@ -391,7 +391,7 @@ class Component(ObservableMixin):
 
             @component.register(
                 "com.example.add",
-                options=RegisterOptions(invoke='round_robin'),
+                options=RegisterOptions(invoke='roundrobin'),
             )
             def add(*args, **kw):
                 print("add({}, {}): event received".format(args, kw))
@@ -780,6 +780,7 @@ class Component(ObservableMixin):
                 # it completely (i.e. until its Deferred fires) and
                 # then disconnect this session
                 def on_join(session, details):
+                    transport.reset()
                     transport.connect_sucesses += 1
                     self.log.debug("session on_join: {details}", details=details)
                     d = txaio.as_future(self._entry, reactor, session)
