@@ -25,9 +25,15 @@
 ###############################################################################
 
 import abc
-from typing import Union
+from typing import Union, Dict, Any, Optional, List, Tuple, Callable
+
+# FIXME: see ISecurityModule.__iter__
+# from collections.abc import Iterator
 
 from autobahn.util import public
+from autobahn.wamp.types import Challenge, SessionDetails, CloseDetails, CallResult, RegisterOptions, \
+    SubscribeOptions, Registration, Subscription, Publication, ComponentConfig, TransportDetails
+from autobahn.wamp.message import Message, Welcome
 
 __all__ = (
     'IObjectSerializer',
@@ -36,8 +42,12 @@ __all__ = (
     'ITransport',
     'ITransportHandler',
     'ISession',
+    'IAuthenticator',
+    'ISigningKey',
+    'IEd25519Key',
+    'IEthereumKey',
+    'ISecurityModule',
     'IPayloadCodec',
-    'ISigningKey'
 )
 
 
@@ -50,8 +60,17 @@ class IObjectSerializer(abc.ABC):
     """
 
     @public
-    @abc.abstractproperty
-    def BINARY(self):
+    @property
+    @abc.abstractmethod
+    def NAME(self) -> str:
+        """
+        Object serializer name (read-only).
+        """
+
+    @public
+    @property
+    @abc.abstractmethod
+    def BINARY(self) -> bool:
         """
         Flag (read-only) to indicate if serializer requires a binary clean
         transport or if UTF8 transparency is sufficient.
@@ -59,28 +78,24 @@ class IObjectSerializer(abc.ABC):
 
     @public
     @abc.abstractmethod
-    def serialize(self, obj):
+    def serialize(self, obj: Any) -> bytes:
         """
         Serialize an object to a byte string.
 
-        :param obj: Object to serialize.
-        :type obj: any (serializable type)
+        :param obj: Object (any serializable type) to serialize.
 
         :returns: Serialized bytes.
-        :rtype: bytes
         """
 
     @public
     @abc.abstractmethod
-    def unserialize(self, payload):
+    def unserialize(self, payload: bytes) -> List[Any]:
         """
-        Unserialize objects from a byte string.
+        Deserialize objects from a byte string.
 
-        :param payload: Objects to unserialize.
-        :type payload: bytes
+        :param payload: Objects to deserialize.
 
-        :returns: List of (raw) objects unserialized.
-        :rtype: list
+        :returns: List of deserialized (raw) objects.
         """
 
 
@@ -91,58 +106,74 @@ class ISerializer(abc.ABC):
     """
 
     @public
-    @abc.abstractproperty
-    def MESSAGE_TYPE_MAP(self):
+    @property
+    @abc.abstractmethod
+    def MESSAGE_TYPE_MAP(self) -> Dict[int, 'IMessage']:
         """
         Mapping of WAMP message type codes to WAMP message classes.
         """
 
     @public
-    @abc.abstractproperty
-    def SERIALIZER_ID(self):
+    @property
+    @abc.abstractmethod
+    def SERIALIZER_ID(self) -> str:
         """
-        The WAMP serialization format ID.
+        The WAMP serialization format ID as used for WebSocket, e.g. ``"json"`` (or ``"json.batched"``) for JSON.
+        """
+
+    @public
+    @property
+    @abc.abstractmethod
+    def RAWSOCKET_SERIALIZER_ID(self) -> int:
+        """
+        The WAMP serialization format ID as used for RawSocket, e.g. ``1`` for JSON.
+        """
+
+    @public
+    @property
+    @abc.abstractmethod
+    def MIME_TYPE(self) -> str:
+        """
+        The WAMP serialization format MIME type, e.g. ``"application/json"`` for JSON.
         """
 
     @public
     @abc.abstractmethod
-    def serialize(self, message):
+    def serialize(self, message: 'IMessage') -> Tuple[bytes, bool]:
         """
-        Serializes a WAMP message to bytes for sending over a transport.
+        Serializes a WAMP message to bytes for sending over a WAMP transport.
 
         :param message: The WAMP message to be serialized.
-        :type message: object implementing :class:`autobahn.wamp.interfaces.IMessage`
 
-        :returns: A pair ``(payload, isBinary)``.
-        :rtype: tuple
+        :returns: A pair ``(payload, is_binary)``.
         """
 
     @public
     @abc.abstractmethod
-    def unserialize(self, payload, isBinary):
+    def unserialize(self, payload: bytes, is_binary: Optional[bool] = None) -> List['IMessage']:
         """
         Deserialize bytes from a transport and parse into WAMP messages.
 
         :param payload: Byte string from wire.
-        :type payload: bytes
 
         :param is_binary: Type of payload. True if payload is a binary string, else
             the payload is UTF-8 encoded Unicode text.
-        :type is_binary: bool
 
-        :returns: List of ``a.w.m.Message`` objects.
-        :rtype: list
+        :returns: List of WAMP messages.
         """
 
 
 @public
 class IMessage(abc.ABC):
     """
+    A WAMP message, e.g. one of the messages defined in the WAMP specification
+    `here <https://wamp-proto.org/_static/gen/wamp_latest_ietf.html#rfc.section.6.5>`_.
     """
 
     @public
-    @abc.abstractproperty
-    def MESSAGE_TYPE(self):
+    @property
+    @abc.abstractmethod
+    def MESSAGE_TYPE(self) -> int:
         """
         WAMP message type code.
         """
@@ -152,29 +183,26 @@ class IMessage(abc.ABC):
     @public
     @staticmethod
     @abc.abstractmethod
-    def parse(wmsg):
+    def parse(wmsg) -> 'IMessage':
         """
         Factory method that parses a unserialized raw message (as returned byte
         :func:`autobahn.interfaces.ISerializer.unserialize`) into an instance
         of this class.
 
         :returns: The parsed WAMP message.
-        :rtype: object implementing :class:`autobahn.wamp.interfaces.IMessage`
         """
 
     @public
     @abc.abstractmethod
-    def serialize(self, serializer):
+    def serialize(self, serializer: ISerializer) -> bytes:
         """
         Serialize this object into a wire level bytes representation and cache
         the resulting bytes. If the cache already contains an entry for the given
         serializer, return the cached representation directly.
 
         :param serializer: The wire level serializer to use.
-        :type serializer: object implementing :class:`autobahn.wamp.interfaces.ISerializer`
 
         :returns: The serialized bytes.
-        :rtype: bytes
         """
 
     @public
@@ -185,6 +213,9 @@ class IMessage(abc.ABC):
         """
 
 
+IMessage.register(Message)
+
+
 @public
 class ITransport(abc.ABC):
     """
@@ -194,7 +225,7 @@ class ITransport(abc.ABC):
 
     @public
     @abc.abstractmethod
-    def send(self, message):
+    def send(self, message: IMessage):
         """
         Send a WAMP message over the transport to the peer. If the transport is
         not open, this raises :class:`autobahn.wamp.exception.TransportLost`.
@@ -203,19 +234,23 @@ class ITransport(abc.ABC):
         has not yet fired, the send will fail immediately.
 
         :param message: The WAMP message to send over the transport.
-        :type message: object implementing :class:`autobahn.wamp.interfaces.IMessage`
-
-        :returns: obj -- A Deferred/Future
         """
 
     @public
     @abc.abstractmethod
-    def isOpen(self):
+    def isOpen(self) -> bool:
         """
         Check if the transport is open for messaging.
 
         :returns: ``True``, if the transport is open.
-        :rtype: bool
+        """
+
+    @public
+    @property
+    @abc.abstractmethod
+    def transport_details(self) -> Optional[TransportDetails]:
+        """
+        Return details about the transport (when the transport is open).
         """
 
     @public
@@ -237,102 +272,45 @@ class ITransport(abc.ABC):
         detected attacks.
         """
 
-    @public
-    @abc.abstractmethod
-    def get_channel_id(self):
-        """
-        Return the unique channel ID of the underlying transport. This is used to
-        mitigate credential forwarding man-in-the-middle attacks when running
-        application level authentication (eg WAMP-cryptosign) which are decoupled
-        from the underlying transport.
-
-        The channel ID is only available when running over TLS (either WAMP-WebSocket
-        or WAMP-RawSocket). It is not available for non-TLS transports (plain TCP or
-        Unix domain sockets). It is also not available for WAMP-over-HTTP/Longpoll.
-        Further, it is currently unimplemented for asyncio (only works on Twisted).
-
-        The channel ID is computed as follows:
-
-           - for a client, the SHA256 over the "TLS Finished" message sent by the client
-             to the server is returned.
-
-           - for a server, the SHA256 over the "TLS Finished" message the server expected
-             the client to send
-
-        Note: this is similar to `tls-unique` as described in RFC5929, but instead
-        of returning the raw "TLS Finished" message, it returns a SHA256 over such a
-        message. The reason is that we use the channel ID mainly with WAMP-cryptosign,
-        which is based on Ed25519, where keys are always 32 bytes. And having a channel ID
-        which is always 32 bytes (independent of the TLS ciphers/hashfuns in use) allows
-        use to easily XOR channel IDs with Ed25519 keys and WAMP-cryptosign challenges.
-
-        WARNING: For safe use of this (that is, for safely binding app level authentication
-        to the underlying transport), you MUST use TLS, and you SHOULD deactivate both
-        TLS session renegotiation and TLS session resumption.
-
-        References:
-
-           - https://tools.ietf.org/html/rfc5056
-           - https://tools.ietf.org/html/rfc5929
-           - http://www.pyopenssl.org/en/stable/api/ssl.html#OpenSSL.SSL.Connection.get_finished
-           - http://www.pyopenssl.org/en/stable/api/ssl.html#OpenSSL.SSL.Connection.get_peer_finished
-
-        :returns: The channel ID (if available) of the underlying WAMP transport. The
-            channel ID is a 32 bytes value.
-        :rtype: binary or None
-        """
-
 
 @public
 class ITransportHandler(abc.ABC):
 
-    @public
-    @abc.abstractproperty
-    def transport(self):
-        """
-        When the transport this handler is attached to is currently open, this property
-        can be read from. The property should be considered read-only. When the transport
-        is gone, this property is set to None.
-        """
-
     @public
     @abc.abstractmethod
-    def onOpen(self, transport):
+    def onOpen(self, transport: ITransport):
         """
         Callback fired when transport is open. May run asynchronously. The transport
         is considered running and is_open() would return true, as soon as this callback
         has completed successfully.
 
         :param transport: The WAMP transport.
-        :type transport: object implementing :class:`autobahn.wamp.interfaces.ITransport`
         """
 
     @public
     @abc.abstractmethod
-    def onMessage(self, message):
+    def onMessage(self, message: IMessage):
         """
         Callback fired when a WAMP message was received. May run asynchronously. The callback
         should return or fire the returned deferred/future when it's done processing the message.
         In particular, an implementation of this callback must not access the message afterwards.
 
         :param message: The WAMP message received.
-        :type message: object implementing :class:`autobahn.wamp.interfaces.IMessage`
         """
 
     @public
     @abc.abstractmethod
-    def onClose(self, wasClean):
+    def onClose(self, wasClean: bool):
         """
         Callback fired when the transport has been closed.
 
         :param wasClean: Indicates if the transport has been closed regularly.
-        :type wasClean: bool
         """
 
 
 # ISession.register collides with the abc.ABCMeta.register method
 class _ABC(abc.ABC):
-    _abc_register = abc.ABC.register
+    abc_register = abc.ABC.register
 
 
 @public
@@ -341,12 +319,31 @@ class ISession(_ABC):
     Interface for WAMP sessions.
     """
 
+    @public
+    @property
     @abc.abstractmethod
-    def __init__(self, config=None):
+    def config(self) -> ComponentConfig:
+        """
+        Configuration for session.
         """
 
-        :param config: Configuration for session.
-        :type config: instance of :class:`autobahn.wamp.types.ComponentConfig`.
+    @public
+    @property
+    @abc.abstractmethod
+    def transport(self) -> Optional[ITransport]:
+        """
+        When the transport this session is attached to is currently open, this property
+        can be read from. The property should be considered read-only. When the transport
+        is gone, this property is set to None.
+        """
+
+    @public
+    @property
+    @abc.abstractmethod
+    def session_details(self) -> Optional[SessionDetails]:
+        """
+        Return details about the session, the same as initially provided to the
+        :meth:`ISession.onJoin` callback on an implementation.
         """
 
     @public
@@ -380,33 +377,32 @@ class ISession(_ABC):
     @public
     @abc.abstractmethod
     def join(self,
-             realm,
-             authmethods=None,
-             authid=None,
-             authrole=None,
-             authextra=None,
-             resumable=None,
-             resume_session=None,
-             resume_token=None):
+             realm: str,
+             authmethods: Optional[List[str]] = None,
+             authid: Optional[str] = None,
+             authrole: Optional[str] = None,
+             authextra: Optional[Dict[str, Any]] = None,
+             resumable: Optional[bool] = None,
+             resume_session: Optional[int] = None,
+             resume_token: Optional[str] = None):
         """
         Attach the session to the given realm. A session is open as soon as it is attached to a realm.
         """
 
     @public
     @abc.abstractmethod
-    def onChallenge(self, challenge):
+    def onChallenge(self, challenge: Challenge) -> str:
         """
         Callback fired when the peer demands authentication.
 
         May return a Deferred/Future.
 
         :param challenge: The authentication challenge.
-        :type challenge: Instance of :class:`autobahn.wamp.types.Challenge`.
         """
 
     @public
     @abc.abstractmethod
-    def onWelcome(self, welcome_msg):
+    def onWelcome(self, welcome: Welcome) -> Optional[str]:
         """
         Callback fired after the peer has successfully authenticated. If
         this returns anything other than None/False, the session is
@@ -414,49 +410,51 @@ class ISession(_ABC):
 
         May return a Deferred/Future.
 
-        :param welcome_msg: The WELCOME message received from the server
-        :type challenge: Instance of :class:`autobahn.wamp.message.Welcome`.
+        .. note::
+            Before we let user code see the session -- that is, before we fire "join"
+            we give authentication instances a chance to abort the session. Usually
+            this would be for "mutual authentication" scenarios. For example, WAMP-SCRAM
+            uses this to confirm the server-signature.
 
-        :return: None, or an error message
+        :param welcome: The WELCOME message received from the server
+
+        :return: None, or an error message (using a fixed error URI
+            ``wamp.error.cannot_authenticate``).
         """
 
     @public
     @abc.abstractmethod
-    def onJoin(self, details):
+    def onJoin(self, details: SessionDetails):
         """
         Callback fired when WAMP session has been established.
 
         May return a Deferred/Future.
 
         :param details: Session information.
-        :type details: Instance of :class:`autobahn.wamp.types.SessionDetails`.
         """
 
     @public
     @abc.abstractmethod
-    def leave(self, reason=None, message=None):
+    def leave(self, reason: Optional[str] = None, message: Optional[str] = None):
         """
         Actively close this WAMP session.
 
         :param reason: An optional URI for the closing reason. If you
-            want to permanently log out, this should be `wamp.close.logout`
-        :type reason: str
+            want to permanently log out, this should be ``wamp.close.logout``.
 
-        :param message: An optional (human readable) closing message, intended for
-                        logging purposes.
-        :type message: str
+        :param message: An optional (human-readable) closing message, intended for
+            logging purposes.
 
         :return: may return a Future/Deferred that fires when we've disconnected
         """
 
     @public
     @abc.abstractmethod
-    def onLeave(self, details):
+    def onLeave(self, details: CloseDetails):
         """
         Callback fired when WAMP session has is closed
 
-        :param details: Close information.
-        :type details: Instance of :class:`autobahn.wamp.types.CloseDetails`.
+        :param details: Close information for session.
         """
 
     @public
@@ -475,21 +473,21 @@ class ISession(_ABC):
 
     @public
     @abc.abstractmethod
-    def is_connected(self):
+    def is_connected(self) -> bool:
         """
         Check if the underlying transport is connected.
         """
 
     @public
     @abc.abstractmethod
-    def is_attached(self):
+    def is_attached(self) -> bool:
         """
         Check if the session has currently joined a realm.
         """
 
     @public
     @abc.abstractmethod
-    def set_payload_codec(self, payload_codec):
+    def set_payload_codec(self, payload_codec: Optional['IPayloadCodec']):
         """
         Set a payload codec on the session. To remove a previously set payload codec,
         set the codec to ``None``.
@@ -498,40 +496,35 @@ class ISession(_ABC):
 
         :param payload_codec: The payload codec that should process application
             payload of the given encoding.
-        :type payload_codec: object
-            implementing :class:`autobahn.wamp.interfaces.IPayloadCodec` or ``None``
         """
 
     @public
     @abc.abstractmethod
-    def get_payload_codec(self):
+    def get_payload_codec(self) -> Optional['IPayloadCodec']:
         """
         Get the current payload codec (if any) for the session.
 
         Payload codecs are used with WAMP payload transparency mode.
 
         :returns: The current payload codec or ``None`` if no codec is active.
-        :rtype: object implementing
-            :class:`autobahn.wamp.interfaces.IPayloadCodec` or ``None``
         """
 
     @public
     @abc.abstractmethod
-    def define(self, exception, error=None):
+    def define(self, exception: Exception, error: Optional[str] = None):
         """
         Defines an exception for a WAMP error in the context of this WAMP session.
 
         :param exception: The exception class to define an error mapping for.
-        :type exception: A class that derives of ``Exception``.
 
         :param error: The URI (or URI pattern) the exception class should be mapped for.
             Iff the ``exception`` class is decorated, this must be ``None``.
-        :type error: str
         """
 
     @public
     @abc.abstractmethod
-    def call(self, procedure, *args, **kwargs):
+    def call(self, procedure: str, *args: Optional[List[Any]], **kwargs: Optional[Dict[str, Any]]) -> \
+            Union[Any, CallResult]:
         """
         Call a remote procedure.
 
@@ -554,21 +547,19 @@ class ISession(_ABC):
         be canceled by canceling the returned Deferred/Future.
 
         :param procedure: The URI of the remote procedure to be called, e.g. ``"com.myapp.hello"``.
-        :type procedure: unicode
 
         :param args: Any positional arguments for the call.
-        :type args: list
 
         :param kwargs: Any keyword arguments for the call.
-        :type kwargs: dict
 
-        :returns: A Deferred/Future for the call result -
-        :rtype: instance of :tx:`twisted.internet.defer.Deferred` / :py:class:`asyncio.Future`
+        :returns: A Deferred/Future for the call result.
         """
 
     @public
     @abc.abstractmethod
-    def register(self, endpoint, procedure=None, options=None, prefix=None, check_types=None):
+    def register(self, endpoint: Union[Callable, Any], procedure: Optional[str] = None,
+                 options: Optional[RegisterOptions] = None, prefix: Optional[str] = None,
+                 check_types: Optional[bool] = None) -> Union[Registration, List[Registration]]:
         """
         Register a procedure for remote calling.
 
@@ -588,23 +579,18 @@ class ISession(_ABC):
         DeferredList or Future is returned that gathers all individual underlying Deferreds/Futures.
 
         :param endpoint: The endpoint called under the procedure.
-        :type endpoint: callable or object
 
         :param procedure: When ``endpoint`` is a callable, the URI (or URI pattern)
            of the procedure to register for. When ``endpoint`` is an object,
            the argument is ignored (and should be ``None``).
-        :type procedure: unicode
 
         :param options: Options for registering.
-        :type options: instance of :class:`autobahn.wamp.types.RegisterOptions`.
-
 
         :param prefix: if not None, this specifies a prefix to prepend
             to all URIs registered for this class. So if there was an
             @wamp.register('method_foo') on a method and
             prefix='com.something.' then a method
             'com.something.method_foo' would ultimately be registered.
-        :type prefix: str
 
         :param check_types: Enable automatic type checking against (Python 3.5+) type hints
             specified on the ``endpoint`` callable. Types are checked at run-time on each
@@ -613,15 +599,14 @@ class ISession(_ABC):
             :class:`autobahn.wamp.protocol.ApplicationSession`. An error
             of type :class:`autobahn.wamp.exception.TypeCheckError` is also raised and
             returned to the caller (via the router).
-        :type check_types: bool
 
         :returns: A registration or a list of registrations (or errors)
-        :rtype: instance(s) of :tx:`twisted.internet.defer.Deferred` / :py:class:`asyncio.Future`
         """
 
     @public
     @abc.abstractmethod
-    def publish(self, topic, *args, **kwargs):
+    def publish(self, topic: str, *args: Optional[List[Any]], **kwargs: Optional[Dict[str, Any]]) -> \
+            Optional[Publication]:
         """
         Publish an event to a topic.
 
@@ -644,21 +629,19 @@ class ISession(_ABC):
           of :class:`autobahn.wamp.exception.ApplicationError`.
 
         :param topic: The URI of the topic to publish to, e.g. ``"com.myapp.mytopic1"``.
-        :type topic: unicode
 
         :param args: Arbitrary application payload for the event (positional arguments).
-        :type args: list
 
         :param kwargs: Arbitrary application payload for the event (keyword arguments).
-        :type kwargs: dict
 
         :returns: Acknowledgement for acknowledge publications - otherwise nothing.
-        :rtype: ``None`` or instance of :tx:`twisted.internet.defer.Deferred` / :py:class:`asyncio.Future`
         """
 
     @public
     @abc.abstractmethod
-    def subscribe(self, handler, topic=None, options=None, check_types=None):
+    def subscribe(self, handler: Union[Callable, Any], topic: Optional[str] = None,
+                  options: Optional[SubscribeOptions] = None, check_types: Optional[bool] = None) -> \
+            Union[Subscription, List[Subscription]]:
         """
         Subscribe to a topic for receiving events.
 
@@ -678,33 +661,30 @@ class ISession(_ABC):
         and a list of Deferreds/Futures is returned that each resolves or rejects as above.
 
         :param handler: The event handler to receive events.
-        :type handler: callable or object
 
         :param topic: When ``handler`` is a callable, the URI (or URI pattern)
            of the topic to subscribe to. When ``handler`` is an object, this
            value is ignored (and should be ``None``).
-        :type topic: unicode
 
         :param options: Options for subscribing.
-        :type options: An instance of :class:`autobahn.wamp.types.SubscribeOptions`.
 
         :param check_types: Enable automatic type checking against (Python 3.5+) type hints
             specified on the ``endpoint`` callable. Types are checked at run-time on each
             invocation of the ``endpoint`` callable. When a type mismatch occurs, the error
             is forwarded to the subscriber code in ``onUserError`` override method of
             :class:`autobahn.wamp.protocol.ApplicationSession`.
-        :type check_types: bool
 
         :returns: A single Deferred/Future or a list of such objects
-        :rtype: instance(s) of :tx:`twisted.internet.defer.Deferred` / :py:class:`asyncio.Future`
         """
 
 
-# experimental authentication API
 class IAuthenticator(abc.ABC):
+    """
+    Experimental authentication API.
+    """
 
     @abc.abstractmethod
-    def on_challenge(self, session, challenge):
+    def on_challenge(self, session: ISession, challenge: Challenge):
         """
         Formulate a challenge response for the given session and Challenge
         instance. This is sent to the server in the AUTHENTICATE
@@ -712,7 +692,7 @@ class IAuthenticator(abc.ABC):
         """
 
     @abc.abstractmethod
-    def on_welcome(self, authextra):
+    def on_welcome(self, authextra: Optional[Dict[str, Any]]) -> Optional[str]:
         """
         This hook is called when the onWelcome/on_welcome hook is invoked
         in the protocol, with the 'authextra' dict extracted from the
@@ -725,18 +705,38 @@ class IAuthenticator(abc.ABC):
 
 @public
 class ISigningKey(abc.ABC):
+    """
+    Interface to an asymmetric verification key, e.g. a WAMP-Cryptosign client or server authentication
+    public key (with Ed25519), or a WAMP-XBR data transaction signature public key or address (with Ethereum).
 
+    The key implementation can use various methods, such as a key read from a file, database table
+    or a key residing in a hardware device.
+    """
+
+    @property
     @abc.abstractmethod
-    def can_sign(self):
+    def key_id(self) -> str:
         """
-        Check if the key can be used to sign.
 
-        :returns: `True`, iff the key can sign.
-        :rtype: bool
+        :return:
+        """
+
+    @property
+    @abc.abstractmethod
+    def key_type(self) -> str:
+        """
+        Type of key and signature scheme, currently one of:
+
+        * ``ed25519``: Ed25519, that is **EdDSA** signing algo with **Curve25519** elliptic curve and **SHA-512** hash,
+                used with WAMP-cryptosign session authentication
+        * ``eth``: Ethereum, that is **ECDSA** signing algo, **secp256k1** elliptic curve and **Keccak-256** hash,
+                used with WAMP-XBR data and transaction signatures
+
+        :return: Key type, one of ``ed25519`` or ``eth``.
         """
 
     @abc.abstractmethod
-    def public_key(self, binary=False) -> Union[str, bytes]:
+    def public_key(self, binary: bool = False) -> Union[str, bytes]:
         """
         Returns the public key part of a signing key or the (public) verification key.
 
@@ -745,15 +745,290 @@ class ISigningKey(abc.ABC):
         """
 
     @abc.abstractmethod
-    def sign(self, data: bytes):
+    def can_sign(self) -> bool:
         """
-        Sign the given data.
+        Check if the key can be used to sign and create new signatures, or only to verify signatures.
+
+        :returns: ``True``, if the key can be used for signing.
+        """
+
+    @abc.abstractmethod
+    def sign(self, data: bytes) -> bytes:
+        """
+        Sign the given data, only available if ``can_sign == True``. This method (always) runs asynchronously.
 
         :param data: The data to be signed.
-        :type data: bytes
 
-        :returns: The signature.
-        :rtype: txaio.Future object that resolves to bytes
+        :return: The signature, that is a future object that resolves to bytes.
+        """
+
+    @abc.abstractmethod
+    def recover(self, data: bytes, signature: bytes) -> bytes:
+        """
+        Recover the signer from the data signed, and the signature given. This method (always) runs asynchronously.
+
+        :param data: The data that was signed.
+        :param signature: The signature over the data.
+
+        :return: The signer public key that signed the data to create the signature given.
+        """
+
+
+@public
+class IEd25519Key(ISigningKey):
+    """
+    Interface to a WAMP-Cryptosign client authentication (or server verification) key.
+    """
+
+    @abc.abstractmethod
+    def sign_challenge(self, challenge: Challenge, channel_id: Optional[bytes] = None,
+                       channel_id_type: Optional[str] = None) -> bytes:
+        """
+        Sign the data from the given WAMP challenge message, and the optional TLS channel ID
+        using this key and return a valid signature that can be used in a WAMP-cryptosign
+        authentication handshake.
+
+        :param challenge: The WAMP challenge message as sent or received during the WAMP-cryptosign
+            authentication handshake. This can be used by WAMP clients to compute the signature
+            returned in the handshake, or by WAMP routers to verify the signature returned by clients,
+            during WAMP-cryptosign client authentication.
+
+        :param channel_id: Optional TLS channel ID. Using this binds the WAMP session authentication
+            to the underlying TLS channel, and thus prevents authentication-forwarding attacks.
+        :param channel_id_type: Optional TLS channel ID type, e.g. ``"tls-unique"``.
+
+        :return: The signature, that is a future object that resolves to bytes.
+        """
+
+    @abc.abstractmethod
+    def verify_challenge(self, challenge: Challenge, signature: bytes, channel_id: Optional[bytes] = None,
+                         channel_id_type: Optional[str] = None) -> bool:
+        """
+        Verify the data from the given WAMP challenge message, and the optional TLS channel ID
+        to be signed by this key.
+
+        :param challenge: The WAMP challenge message as sent or received during the WAMP-cryptosign
+            authentication handshake. This can be used by WAMP clients to compute the signature
+            returned in the handshake, or by WAMP routers to verify the signature returned by clients,
+            during WAMP-cryptosign client authentication.
+
+        :param channel_id: Optional TLS channel ID. Using this binds the WAMP session authentication
+            to the underlying TLS channel, and thus prevents authentication-forwarding attacks.
+        :param channel_id_type: Optional TLS channel ID type, e.g. ``"tls-unique"``.
+
+        :return: Returns ``True`` if the signature over the data matches this key.
+        """
+
+
+@public
+class IEthereumKey(ISigningKey):
+    """
+    Interface to an Ethereum signing (or transaction verification) key, used for WAMP-XBR transaction
+    signing (or verification).
+    """
+
+    @abc.abstractmethod
+    def address(self, binary: bool = False) -> Union[str, bytes]:
+        """
+        Returns the Ethereum (public) address of the key (which is derived from
+        the public key).
+
+        :param binary: Return address as 160 bits (20 bytes) binary instead of
+            the ``0x`` prefixed hex, check-summed address as a string.
+        :return: The address in hex or byte encoding.
+        """
+
+    @abc.abstractmethod
+    def sign_typed_data(self, data: Dict[str, Any]) -> bytes:
+        """
+        Sign the given typed data according to `EIP712 <https://eips.ethereum.org/EIPS/eip-712>`_
+        and create an Ethereum signature.
+
+        :param data: The data to be signed. This must follow EIP712.
+
+        :return: The signature, that is a future object that resolves to bytes.
+        """
+
+    @abc.abstractmethod
+    def verify_typed_data(self, data: Dict[str, Any], signature: bytes) -> bool:
+        """
+        Verify the given typed data according to `EIP712 <https://eips.ethereum.org/EIPS/eip-712>`_
+        to be signed by this key.
+
+        :param data: The data to be signed. This must follow EIP712.
+        :param signature: The signature to be verified.
+
+        :return: Returns ``True`` if the signature over the data matches this key.
+        """
+
+
+@public
+class ISecurityModule(abc.ABC):
+    """
+    Interface for key security modules, which
+
+    * include filesystem and HSM backed persistent key implementations, and
+    * provides secure key signature generation and verification with
+    * two key types and signature schemes
+
+    The two key types and signature schemes support WAMP-cryptosign based authentication
+    for WAMP sessions, and WAMP-XBR based signed transactions and data encryption.
+
+    References:
+
+    * `SE050 APDU Specification (AN12413) <https://www.nxp.com/docs/en/application-note/AN12413.pdf>`_
+    * https://neuromancer.sk/std/secg/secp256r1
+    * https://neuromancer.sk/std/secg/secp256k1
+    * https://asecuritysite.com/curve25519/eddsa2
+    * https://asecuritysite.com/secp256k1/ecdsa
+    * https://safecurves.cr.yp.to/
+    * https://www.ietf.org/rfc/rfc3279.txt
+    * https://crypto.stackexchange.com/questions/70927/naming-convention-for-nist-elliptic-curves-in-openssl
+    * https://www.johndcook.com/blog/2018/08/21/a-tale-of-two-elliptic-curves/
+    """
+
+    @abc.abstractmethod
+    def open(self):
+        """
+        Open this security module. This method (always) runs asynchronously.
+        """
+
+    @abc.abstractmethod
+    def close(self):
+        """
+        Close this security module. This method (always) runs asynchronously.
+        """
+
+    @abc.abstractmethod
+    def is_open(self) -> bool:
+        """
+        Check if the security module is currently opened. Security module operations
+        can only be run when the module is opened.
+
+        :return: Flag indicating whether the security module is currently opened.
+        """
+
+    @abc.abstractmethod
+    def can_lock(self) -> bool:
+        """
+        Flag indicating whether this security module can be locked, e.g. by a
+        user passphrase or PIN.
+
+        :return: Flag indicating whether the security module can be locked/unlocked at all.
+        """
+
+    @abc.abstractmethod
+    def is_locked(self) -> bool:
+        """
+        Check if this security module is currently locked.
+
+        :return: Flag indicating whether the security module is currently locked.
+        """
+
+    @abc.abstractmethod
+    def lock(self):
+        """
+        Lock this security module. This method (always) runs asynchronously.
+        """
+
+    @abc.abstractmethod
+    def unlock(self):
+        """
+        Unlock this security module. This method (always) runs asynchronously.
+        """
+
+    @abc.abstractmethod
+    def random(self, octets: int) -> bytes:
+        """
+        Generate random bytes within the security module.
+
+        :param octets: Number of bytes (octets) to generate.
+
+        :return: Random bytes, generated within the security module, e.g. in a HW RNG.
+        """
+
+    @abc.abstractmethod
+    def current(self, counter_id: str) -> int:
+        """
+        Return current value of the given persistent counter.
+
+        :param counter_id: The ID of the counter to access.
+
+        :return: Current value of counter, or ``0`` to indicate the counter does not
+            exists (was never incremented).
+        """
+
+    @abc.abstractmethod
+    def next(self, counter_id: str) -> int:
+        """
+        Increment the given persistent counter and return the new value.
+
+        :param counter_id:
+
+        :param counter_id: The ID of the counter to access.
+
+        :return: New value of counter, e.g. ``1`` once a counter was first incremented.
+        """
+
+    @abc.abstractmethod
+    def create(self, key_type: str) -> str:
+        """
+        Create a new public-private asymmetric key pair, stored within the security module.
+
+        :param key_type: Type of key to generate, e.g. ``ed25519`` or ``eth``.
+
+        :return: ID of new key.
+        """
+
+    @abc.abstractmethod
+    def delete(self, key_id: str):
+        """
+        Delete an existing key pair stored within the security module.
+
+        :param key_id: ID of key to delete.
+        """
+
+    @abc.abstractmethod
+    def __len__(self) -> int:
+        """
+        Get number of key pairs currently stored within the security module.
+
+        :return: Current number of persistent keys.
+        """
+
+    # FIXME: the following works on CPy 3.9+, but fails on CPy 3.7 and PyPy 3.8
+    #   AttributeError: type object 'Iterator' has no attribute '__class_getitem__'
+    #   See also:
+    #       - https://docs.python.org/3/library/abc.html#abc.ABCMeta.__subclasshook__
+    #       - https://docs.python.org/3/library/stdtypes.html#container.__iter__
+    #
+    # @abc.abstractmethod
+    # def __iter__(self) -> Iterator[str]:
+    #     """
+    #     Return an iterator object over all key accessible in this security module.
+    #
+    #     :return:
+    #     """
+
+    @abc.abstractmethod
+    def __contains__(self, key_id: str) -> bool:
+        """
+        Check if the security module, acting as a key container, contains a persistent
+        key with the given ID.
+
+        :param key_id: ID of key to check.
+
+        :return: ``True`` if a key with given ID exists.
+        """
+
+    @abc.abstractmethod
+    def __getitem__(self, key_id: str) -> Union[IEd25519Key, IEthereumKey]:
+        """
+        Get a key from the security module given the key ID.
+
+        :param key_id: ID of key to get.
+
+        :return: The key, either a :class:`IEd25519Key` or :class:`IEthereumKey` instance.
         """
 
 
@@ -814,8 +1089,8 @@ class IPayloadCodec(abc.ABC):
             the payload is to be encoded (eg topic or procedure).
         :type uri: str
 
-        :param payload: The encoded application payload to be decoded.
-        :type payload: instance of :class:`autobahn.wamp.types.EncodedPayload`
+        :param encoded_payload: The encoded application payload to be decoded.
+        :type encoded_payload: instance of :class:`autobahn.wamp.types.EncodedPayload`
 
         :returns: A tuple with the decoded positional and keyword-based
             application payload: ``(uri, args, kwargs)``
