@@ -24,6 +24,16 @@
 #
 ###############################################################################
 
+import traceback
+
+try:
+    # PyPy 7.3.3 lacks this
+    from _hashlib import HASH  # noqa: F401
+except ImportError:
+    # monkey patch it:
+    import _hashlib
+    _hashlib.HASH = _hashlib.Hash
+
 try:
     from mnemonic import Mnemonic
     from autobahn.xbr._mnemonic import mnemonic_to_private_key
@@ -33,10 +43,12 @@ try:
     # https://github.com/ethereum/eth-abi/pull/88
     from eth_abi import abi
 
+    import eth_account
+
     from autobahn.xbr._abi import XBR_TOKEN_ABI, XBR_NETWORK_ABI, XBR_MARKET_ABI, XBR_CATALOG_ABI, XBR_CHANNEL_ABI  # noqa
     from autobahn.xbr._abi import XBR_DEBUG_TOKEN_ADDR, XBR_DEBUG_NETWORK_ADDR, XBR_DEBUG_MARKET_ADDR, XBR_DEBUG_CATALOG_ADDR, XBR_DEBUG_CHANNEL_ADDR  # noqa
     from autobahn.xbr._abi import XBR_DEBUG_TOKEN_ADDR_SRC, XBR_DEBUG_NETWORK_ADDR_SRC, XBR_DEBUG_MARKET_ADDR_SRC, XBR_DEBUG_CATALOG_ADDR_SRC, XBR_DEBUG_CHANNEL_ADDR_SRC  # noqa
-    from autobahn.xbr._interfaces import IMarketMaker, IProvider, IConsumer, ISeller, IBuyer  # noqa
+    from autobahn.xbr._interfaces import IMarketMaker, IProvider, IConsumer, ISeller, IBuyer, IDelegate  # noqa
     from autobahn.xbr._util import make_w3, pack_uint256, unpack_uint256  # noqa
 
     from autobahn.xbr._eip712_member_register import sign_eip712_member_register, recover_eip712_member_register  # noqa
@@ -59,6 +71,11 @@ try:
 
     from autobahn.xbr._config import load_or_create_profile, UserConfig, Profile  # noqa
 
+    from autobahn.xbr._schema import FbsSchema, FbsObject, FbsType, FbsRPCCall, FbsEnum, FbsService, FbsEnumValue, \
+        FbsAttribute, FbsField, FbsRepository  # noqa
+
+    from autobahn.xbr._wallet import stretch_argon2_secret, expand_argon2_secret, pkm_from_argon2_secret  # noqa
+
     HAS_XBR = True
 
     if not hasattr(abi, 'collapse_type'):
@@ -225,29 +242,59 @@ try:
         Actor is both a XBR Provider and XBR Consumer.
         """
 
-    def generate_seedphrase(strength=128, language='english'):
+    def generate_seedphrase(strength=128, language='english') -> str:
         """
         Generate a new BIP-39 mnemonic seed phrase for use in Ethereum (Metamask, etc).
 
+        See:
+        * https://github.com/bitcoin/bips/blob/master/bip-0039.mediawiki
+        * https://github.com/trezor/python-mnemonic
+
         :param strength: Strength of seed phrase in bits, one of the following ``[128, 160, 192, 224, 256]``,
             generating seed phrase of 12 - 24 words inlength.
 
         :return: Newly generated seed phrase (in english).
-        :rtype: string
         """
         return Mnemonic(language).generate(strength)
 
-    def check_seedphrase(seedphrase, language='english'):
+    def check_seedphrase(seedphrase: str, language: str = 'english'):
+        """
+        Check a BIP-39 mnemonic seed phrase.
+
+        :param seedphrase: The BIP-39 seedphrase from which to derive the account.
+        :param language: The BIP-39 user language to generate the seedphrase for.
+        :return:
+        """
         return Mnemonic(language).check(seedphrase)
 
-    def account_from_seedphrase(seephrase, index=0):
+    def account_from_seedphrase(seedphrase: str, index: int = 0) -> eth_account.account.Account:
+        """
+        Create an account from the given BIP-39 mnemonic seed phrase.
+
+        :param seedphrase: The BIP-39 seedphrase from which to derive the account.
+        :param index: The account index in account hierarchy defined by the seedphrase.
+        :return: The new Eth account object
+        """
         from web3.auto import w3
 
         derivation_path = "m/44'/60'/0'/0/{}".format(index)
-        key = mnemonic_to_private_key(seephrase, str_derivation_path=derivation_path)
+        key = mnemonic_to_private_key(seedphrase, str_derivation_path=derivation_path)
         account = w3.eth.account.privateKeyToAccount(key)
         return account
 
+    def account_from_ethkey(ethkey: bytes) -> eth_account.account.Account:
+        """
+        Create an account from the private key seed.
+
+        :param ethkey: The Ethereum private key seed (32 octets).
+        :return: The new Eth account object
+        """
+        from web3.auto import w3
+
+        assert len(ethkey) == 32
+        account = w3.eth.account.privateKeyToAccount(ethkey)
+        return account
+
     ASCII_BOMB = r"""
             _ ._  _ , _ ._
             (_ ' ( `  )_  .__)
@@ -330,10 +377,27 @@ try:
         'IConsumer',
         'ISeller',
         'IBuyer',
+        'IDelegate',
+
+        'FbsSchema',
+        'FbsService',
+        'FbsType',
+        'FbsObject',
+        'FbsEnum',
+        'FbsEnumValue',
+        'FbsRPCCall',
+        'FbsAttribute',
+        'FbsField',
+        'FbsRepository',
+
+        'stretch_argon2_secret',
+        'expand_argon2_secret',
+        'pkm_from_argon2_secret',
     )
 
 except (ImportError, FileNotFoundError) as e:
     import sys
+    traceback.print_tb(e.__traceback__, file=sys.stderr)
     sys.stderr.write(str(e))
     sys.stderr.flush()
     HAS_XBR = False
